/*
 * find_star.c : automatically find stars 
 *
 * 2003.02.11 -
 * Time-stamp: <2004-12-23 12:29:17 dimm>
 */

#include <stdlib.h>
#include <math.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "seemon2.h"

/* get star position in finder image
 * using socket comm.
 * v0.4.8b3 2003.09.09
 */
int get_finder_starpos(void) {
  int ret;

  ret = read_finder_coords(); /* in socket_client.c */
  if (ret != EXIT_SUCCESS) {
    printf("ERROR: failed reading stellar position in finder.\n");
    fprintf(fout, "ERROR: failed reading stellar position in finder.\n");
    return(EXIT_FAILURE);
  } else {
    //    #ifdef DEBUG
    //    printf("#finder position: %d %d count: %d\n", 
    //	   mseemon.finder_star_center.cx,
    //	   mseemon.finder_star_center.cy,
    //	   mseemon.finder_peakval);
    //    #endif

    /* check obtained coords are normal */
    /* added v0.5.11 20041102 */
    if (mseemon.finder_star_center.cx > 0 &&
	mseemon.finder_star_center.cy > 0 &&
	mseemon.finder_star_center.cx < 1000 &&
	mseemon.finder_star_center.cy < 1000) {
      return(EXIT_SUCCESS);
    } else {
      /* invalid position */
      printf("# Invalid star position in finder.\n");
      if (fout != NULL) {
	fprintf(fout, "# Invalid star position in finder.\n");
      }
      return(EXIT_FAILURE);
    }
  }

  return(EXIT_FAILURE);
}

/* judge if it is cloudy 
 *  from peak count of finder.
 * return -1 if cloudy (peak < 100)
 *        0(EXIT_SUCCESS) if not (peak >= 100)
 *        1(EXIT_FAILURE) if fails
 *
 * v0.5.2 2003.10.02
 */
int is_cloudy(void) {
  int ret;

  ret = read_finder_coords(); /* in socket_client.c */
  if (ret != EXIT_SUCCESS) {
    printf("ERROR: failed reading stellar position in finder.\n");
    fprintf(fout, "# ERROR: failed reading stellar position in finder.\n");
    return(EXIT_FAILURE);
  } else {
    //    #ifdef DEBUG
    //    printf("#finder position: %d %d count: %d\n", 
    //	   mseemon.finder_star_center.cx,
    //	   mseemon.finder_star_center.cy,
    //	   mseemon.finder_peakval);
    //    #endif
    if (mseemon.finder_peakval < FNDR_MIN_PEAK) {
      /* skip if finder crushes, added v0.6.7 20041223 iwata */
      if (mseemon.finder_star_center.cx > 0 && 
	  mseemon.finder_star_center.cy > 0) {
	/* it seems cloudy .. */
	return(-1);
      } else {
	#ifdef DEBUG
	printf("# finder seems to be temporally stopping.\n");
	#endif
      }
    }
  }
  return(EXIT_SUCCESS);
}

/* find stars in image
 * 2003.05.29 update
 *
 * assume two star images allign vertically
 */
int get_stars(void) {
  int tx, ty, sx, sy, i, j, k, cx, cy;
  int flg, cx2;
  int maxx, maxy, maxval;
  int maxx2, maxy2, maxval2;
  int average;
  guchar *pix;
  pixcoord swp;

  /* steps */
  sx = 30;
  sy = 30;
  /* cell size */
  tx = mseemon.star_rect[0].cx;
  ty = mseemon.star_rect[0].cy;

  pix = graybuf;

  #ifdef DEBUG
  printf("now start finding.\n");
  #endif

  maxval = -1;
  
  /* sum values of pixels in each cell
   * and find a cell with highest value
   */
  cx = 0;
  cy = 0;
  flg = 1;

  while (flg) {
    average = 0;
    for (i = 0; i < ty; i++) {
      for (j = 0; j < tx; j++) {
	average += pix[(cy + i) * wid + cx + j];
      }
    }

    if (average > maxval) {
      maxx = cx;
      maxy = cy;
      maxval = average;
    }

    /* increment */
    cx += sx;
    if (cx > wid - tx) {
      cx = 0;
      cy += sy;
      if (cy > hgt - ty) {
	flg = 0;
      }
    }
  }

  /* find second highest cell 
   * along vertical directions 
   * of the highest cell
   */
  maxval2 = -1;
  /* downward */
  cx = maxx;
  cy = maxy - ty;
  flg = 1;

  while (flg) {
    if (cy < 0) break;

    average = 0;
    for (i = 0; i < ty; i++) {
      for (j = 0; j < tx; j++) {
	average += pix[(cy + i) * wid + cx + j];
      }
    }

    if (average > maxval2) {
      maxx2 = cx;
      maxy2 = cy;
      maxval2 = average;
    }

    /* decrement */
    cy -= sy;
  }

  /* upward */
  cx = maxx;
  cy = maxy + ty;
  flg = 1;

  while (flg) {
    if (cy > hgt - ty) break;

    average = 0;
    for (i = 0; i < ty; i++) {
      for (j = 0; j < tx; j++) {
	average += pix[(cy + i) * wid + cx + j];
      }
    }

    if (average > maxval2) {
      maxx2 = cx;
      maxy2 = cy;
      maxval2 = average;
    }

    /* increment */
    cy += sy;
  }

  /*
  #ifdef DEBUG
  printf("max1: %3d,%3d %d\n", maxx, maxy, maxval);
  printf("max2: %3d,%3d %d\n", maxx2, maxy2, maxval2);
  #endif
  */

  /* set star positions */ 
  mseemon.star_center[0].cx = maxx + tx/2;
  mseemon.star_center[0].cy = maxy + ty/2;

  mseemon.star_center[1].cx = maxx2 + tx/2;
  mseemon.star_center[1].cy = maxy2 + ty/2;

  /* keep star 0 is lower than star 1 */
  if (mseemon.star_center[1].cy < mseemon.star_center[0].cy) {
    swp = mseemon.star_center[1];
    mseemon.star_center[1] = mseemon.star_center[0];
    mseemon.star_center[0] = swp;
  }

  /* set star 0 slightly lower, 
   * and star 1 slightly upper 
   * (for avoiding confliction) */
  //  mseemon.star_center[0].cy -= 5;
  //  mseemon.star_center[1].cy += 5;

  /* v.0.5.3 */
  flg_set_rect[0] = 1;
  flg_set_rect[1] = 1;

  /* success */
  return(EXIT_SUCCESS);
}

/* prev. version
int get_stars(void) {
  int *cell;
  int tx, ty, sx, sy, i, j, k, cx, cy;
  int flg, cx2;
  int max1, max2, val1, val2;
  int average;
  guchar *pix;
  pixcoord swp;

  sx = mseemon.star_rect[0].cx - 5;
  sy = mseemon.star_rect[0].cy - 5;
  tx = wid / sx;
  ty = hgt / sy;

  /* memory allocation for cells 
  if ((cell = (int*)malloc(sizeof(int)*tx*ty)) == NULL) {
    printf("failed in allocating memory for image cell boxes in get_stars().\n");
    return(0);
  }

  pix = graybuf;

  /* init 
  for (i=0; i<ty; i++) {
    for (j=0; j<tx; j++) {
      cell[i*tx+j] = 0;
    }
  }

  printf("now start finding.\n");

  /* sum values of pixels in each cell 
  cx = 0;
  cy = 0;
  flg = 1;
  cx2 = 0;
  i = 0;
  j = 0;
  average = 0;
  while (flg) {
    cell[i*tx+j] += (int)(*pix);
    average += (int)(*pix);
    pix++;
    cx++;
    cx2++;
    if (cx >= sx) {
      cx = 0;
      j++;
      if (j >= tx) {
	j = 0;
	for (k = 0; k < (wid - cx2); k++) {
	  pix++;
	}
	cx2 = 0;
	cy++;
	if (cy >= sy) {
	  cy = 0;
	  i++;
	  if (i >= ty) {
	    flg = 0;
	  }
	}
      }
    }
  }

  average /= tx*ty;
  #ifdef DEBUG
  printf("average is %d.\n");
  #endif

  for (i = 0; i < tx*ty; i++) {
    cell[i] -= average;
  }

  /* find highest and the second highest cells 
  val1 = 0;
  val2 = 0;
  cx = 0;
  for (i = 0; i < tx*ty; i++) {
    #ifdef DEBUG
    printf("%4d ", cell[i]);
    cx++;
    if (cx >= tx) {
      printf("\n");
      cx = 0;
    }
    #endif

    if (cell[i] > val1) {
      max1 = i;
      val1 = cell[i];
    } else if (cell[i] > val2) {
      max2 = i;
      val2 = cell[i];
    }
  }

  /* set star positions  
  cy = max1 / tx;
  cx = max1 - tx * cy;
  mseemon.star_center[0].cx = (int)((double)cx+0.5)*sx;
  mseemon.star_center[0].cy = (int)((double)cy+0.5)*sy;

  cy = max2 / tx;
  cx = max2 - tx * cy;
  mseemon.star_center[1].cx = (int)((double)cx+0.5)*sx;
  mseemon.star_center[1].cy = (int)((double)cy+0.5)*sy;

  /* keep star 0 is lower than star 1 
  if (mseemon.star_center[1].cy < mseemon.star_center[0].cy) {
    swp = mseemon.star_center[1];
    mseemon.star_center[1] = mseemon.star_center[0];
    mseemon.star_center[0] = swp;
  }

  /* set star 0 slightly lower, 
   * and star 1 slightly upper 
   * (for avoiding confliction)
  mseemon.star_center[0].cy -= 5;
  mseemon.star_center[1].cy += 5;

  /* success 
  return(1);
}
*/
