/*  
 * Telescope Control Protocol for the Meade LX200 series
 * John Kielkopf (kielkopf@louisville.edu)
 *
 * 17 February 2002 -- Version 1.00
 *
 * These routines originated with lx200.c by
 * Ken Shouse (kshouse@econtrols.com)
 * as modified by Carlos Guirao (cguirao@eso.org).
 *
 * This version supports XmTel, a simple graphical
 * universal telescope control panel linked with a device library.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <math.h>
#include "protocol.h"

#define NULL_PTR(x) (x *)0

#ifndef PI
#define	PI		3.141592653589793
#endif


void ConnectTel(char *port);
void DisconnectTel(void);
int  CheckConnectTel(void);

void SetRate(int newRate);
void SetRateCenter0(void);  /* added by iwata. v0.5.0 2003.09.11 */
void SetLimits(double limitLower, double limitHigher);
void StartSlew(int direction);
void StopSlew(int direction);
void StopSlewAll(void); /* added by iwata. v0.4.8 2003.09.09 */
double GetRA(void);
double GetDec(void);
int  SlewToCoords(double newRA, double newDec);
int  SyncToCoords(double newRA, double newDec);
int  CheckCoords(double desRA, double desDec);
void StartFocusPos(void); /* added by iwata. v0.5.0 2003.09.11 */
void StartFocusNeg(void); /* added by iwata. v0.5.0 2003.09.11 */
void StopFocus(void);     /* added by iwata. v0.5.0 2003.09.11 */

static int TelPortFD;
static int TelConnectFlag = 0;

typedef fd_set telfds;

static int readn(int fd, char *ptr, int nbytes, int sec);
static int writen(int fd, char *ptr, int nbytes);
static int telstat(int fd,int sec,int usec);


int CheckConnectTel(void)
{
  return TelConnectFlag;
}



void ConnectTel(char *port)
{
  struct termios tty;
  char tempChar;
  char returnStr[128];
  int numRead;

  /* printf("Connect port=%s\n",port); */

  if(TelConnectFlag != 0)
    return;

  /* first, do the connect */
  TelPortFD = open(port,O_RDWR);
  if(TelPortFD == -1) {
    printf("TelPortFD is -1.\n");
    return;
  }

  tcgetattr(TelPortFD,&tty);
  cfsetospeed(&tty, (speed_t) B9600);
  cfsetispeed(&tty, (speed_t) B9600);
  tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;
  tty.c_iflag =  IGNBRK;
  tty.c_lflag = 0;
  tty.c_oflag = 0;
  tty.c_cflag |= CLOCAL | CREAD;
  tty.c_cc[VMIN] = 1;
  tty.c_cc[VTIME] = 5;
  tty.c_iflag &= ~(IXON|IXOFF|IXANY);
  tty.c_cflag &= ~(PARENB | PARODD);
  tcsetattr(TelPortFD, TCSANOW, &tty);

  /* flush the input (read) buffer */
  tcflush(TelPortFD,TCIOFLUSH);

  /* send a # to clean up */
  writen(TelPortFD,"#",1);

  /* ask for local time to test connection */
  writen(TelPortFD,"#:GL#",5);
  numRead=readn(TelPortFD,returnStr,9,1);
  returnStr[numRead] = '\0';

  /* diagnostic test of local time */
  printf("LocalTime: %d read %s\n",numRead,returnStr);

  if (numRead == 9) 
  {
    TelConnectFlag = 1;
  
/* Check precision on RA and toggle to high precision if needed */

    writen(TelPortFD,"#:GR#",5);
    numRead = readn(TelPortFD,returnStr,9,1);       
    if (numRead < 9)
    {
      writen(TelPortFD,"#:U#",4);
      usleep(250000.);
    }            
  }  
}

void SetRate(int newRate)
{
  if(newRate == SLEW) 
    writen(TelPortFD,"#:RS#",5);
  else if(newRate == FIND) 
    writen(TelPortFD,"#:RM#",5);
  else if(newRate == CENTER) 
    writen(TelPortFD,"#:RC#",5);
  else if(newRate == GUIDE) 
    writen(TelPortFD,"#:RG#",5);
}

/* added by iwata. v0.5.0 2003.09.11 */
void SetRateCenter0(void) {
  writen(TelPortFD, "#:RC0#",6);
}

void StartSlew(int direction)
{
  if(direction == NORTH)
    writen(TelPortFD,"#:Mn#",5);
  else if(direction == EAST)
    writen(TelPortFD,"#:Me#",5);
  else if(direction == SOUTH)
    writen(TelPortFD,"#:Ms#",5);
  else if(direction == WEST)
    writen(TelPortFD,"#:Mw#",5);
}

void StopSlew(int direction)
{
  if(direction == NORTH)
    writen(TelPortFD,"#:Qn#",5);
  else if(direction == EAST)
    writen(TelPortFD,"#:Qe#",5);
  else if(direction == SOUTH)
    writen(TelPortFD,"#:Qs#",5);
  else if(direction == WEST)
    writen(TelPortFD,"#:Qw#",5);
}

/* added by iwata. v0.4.8 2003.09.09 */
void StopSlewAll(void)
{
  writen(TelPortFD,"#:Q#",4);
}

void DisconnectTel(void)
{
  /* printf("DisconnectTel\n"); */
  if(TelConnectFlag == 1)
    close(TelPortFD);
  TelConnectFlag = 0;
}


/* Read the telescope right ascension */

double GetRA(void)
{
  char  returnStr[10];
  double  raHr,raMin,raSec,returnRA;
  int  numRead;


  writen(TelPortFD,"#:GR#",5);
  numRead=readn(TelPortFD,returnStr,9,1);

        
/*   In low precision mode the LX200 returns hh:mm.m# which would be handled  */
/*   by the following code.  Most LX200's support high precision.             */
/*   ConnectTel tests the precision and tries to toggle to high precision.    */
/*   If that should not work with your version of the LX200 ROM, try          */
/*   uncommenting the following, and commenting out the high precision        */
/*   parsing routines.                                                        */
    
/*   returnStr[8] = '\0';                                                     */
/*   printf("numRead = %d\n RA=%s\n", numRead, returnStr);                    */ 
/*   sscanf(returnStr,"%lf:%lf",&raHr,&raMin);                                */
/*   raSec=0.;                                                                */

  returnStr[8] = '\0';                                       
  sscanf(returnStr,"%lf:%lf:%lf",&raHr,&raMin,&raSec);                  
  returnRA = raHr + raMin/60.0 + raSec/3600.0;

  return returnRA;
}


/* Read the telescope declination */

double GetDec(void)
{
  char returnStr[10];
  double decDeg,decMin,decSec,returnDec;
  int numRead;


  writen(TelPortFD,"#:GD#",5);
  numRead=readn(TelPortFD,returnStr,10,1);


/*   In low precision mode the telescope returns dd:mm which would be handled */
/*   by the following code.  Most LX200's support high precision, so          */
/*   ConnectTel tests the precision and tries to toggle to high precision.    */
/*   If that should not work with your version of the LX200 ROM, try          */
/*   uncommenting the following, and commenting out the high precision        */
/*   parsing routines.                                                        */

/*   Since the LX200 may return something other than ':' as a delimiter,      */ 
/*   we parse each piece.                                                     */

/*   returnStr[3] = returnStr[6] = returnStr[9]='\0';                         */
/*   sscanf(returnStr,"%lf",&decDeg);                                         */
/*   sscanf(returnStr+4,"%lf#",&decMin);                                      */
/*   decSec=0.;                                                               */

  returnStr[3] = returnStr[6] = returnStr[9]='\0';  

  sscanf(returnStr,"%lf",&decDeg);                  
  sscanf(returnStr+4,"%lf:",&decMin);               
  sscanf(returnStr+7,"%lf",&decSec);
  if(decDeg >= 0.0)
    returnDec = decDeg + decMin/60.0 + decSec/3600.0;
  else
    returnDec = decDeg - decMin/60.0 - decSec/3600.0;

  return returnDec;
}

/*  Reset telescope coordinates to new coordinates */

int SyncToCoords(double newRA, double newDec)
{
  int raHr, raMin, raSec;
  int decDeg, decMin, decSec;
  double dms, hms, ms;
  char outputStr[32], inputStr[2048];
  int returnVal;

  /* In high precision mode, we must send hh:mm:ss and sdd:mm:ss */
  /* Allow for truncation if input is in seconds of arc */

  hms = newRA;
  hms = hms + 0.5/3600.;
  raHr = (int) hms;
  ms = hms - (double) raHr;
  ms = ms*60.;
  raMin = (int) ms;
  raSec = (int) 60.*(ms - (double) raMin);

  sprintf(outputStr,"#:Sr %02d:%02d:%02d#\n",raHr,raMin,raSec);
  
  writen(TelPortFD,outputStr,15);
  if (readn(TelPortFD,inputStr,1,1) != 1 )
  { 
    fprintf(stderr,"Unable to set RA in SyncToCoords\n"); 
    return -1;
  }
  if(inputStr[0] != '1')
  {
    fprintf(stderr,"Error setting object RA in SyncToCoords.\n");
    fprintf(stderr,"Tried to send command string: %s\n",outputStr);
    fprintf(stderr,"Received string %s\n",inputStr);  
  }

  if(newDec >= 0.0)
  {
    dms = newDec;
    dms = dms + 0.5/3600.;
    decDeg = (int) dms;
    ms = dms - (double) decDeg;
    ms = ms*60.;
    decMin = (int) ms;
    decSec = (int) 60.*(ms - (double) decMin);
    sprintf(outputStr,"#:Sd +%02d:%02d:%02d#\n",decDeg,decMin,decSec);
  }
  else
  {
    dms = -newDec;
    dms = dms + 0.5/3600.;
    decDeg = (int) dms;
    ms = dms - (double) decDeg;
    ms = ms*60.;
    decMin = (int) ms;
    decSec = (int) 60.*(ms - (double) decMin);
    sprintf(outputStr,"#:Sd -%02d:%02d:%02d#\n",decDeg,decMin,decSec);
  }

  writen(TelPortFD,outputStr,16);
  if (readn(TelPortFD,inputStr,1,1) != 1 ) 
  {
    fprintf(stderr,"Unable to set Dec in SyncToCoords.\n");
    return -1;
  }

  if(inputStr[0] != '1') {
    fprintf(stderr,"Error setting object Dec in SyncToCoords.\n");
    fprintf(stderr,"Tried to send command string: %s\n",outputStr);
    fprintf(stderr,"Received string: %s\n",inputStr);
  }

  writen(TelPortFD,"#:CM#",5);

  /* Look for '#' */

  for (;;) 
  {
    if ( readn(TelPortFD,inputStr,1,1) ) 
    {
      if (inputStr[0] == '#') break;
    }
    else 
    { 
      fprintf(stderr,"No acknowledgement from telescope after SyncToCoords.\n");
      return -1; 
    }
  }
}


int SlewToCoords(double newRA, double newDec)
{
  int raHr, raMin, raSec;
  int decDeg, decMin, decSec;
  double dms, hms, ms;
  char outputStr[32], inputStr[2048];
  int returnVal;

  /* In high precision mode, we must send hh:mm:ss and sdd:mm:ss */
  /* Allow for truncation if input is in seconds of arc */

  hms = newRA;
  hms = hms + 0.5/3600.;
  raHr = (int) hms;
  ms = hms - (double) raHr;
  ms = ms*60.;
  raMin = (int) ms;
  raSec = (int) 60.*(ms - (double) raMin);

  sprintf(outputStr,"#:Sr %02d:%02d:%02d#\n",raHr,raMin,raSec);
 
  writen(TelPortFD,outputStr,15);
  //  if (readn(TelPortFD,inputStr,1,1) != 1 ) exit(0); 
  /* modified 20041209 v0.6.5 iwata */
  if (readn(TelPortFD,inputStr,1,1) != 1 ) return(-1);
  if(inputStr[0] != '1')
  {
    fprintf(stderr,"Error setting object RA in SlewToCoords!\n");
    fprintf(stderr,"Tried to send command string %s\n",outputStr);
    fprintf(stderr,"Received string %s\n",inputStr); 
  }

  if(newDec >= 0.0)
  {
    dms = newDec;
    dms = dms + 0.5/3600.;
    decDeg = (int) dms;
    ms = dms - (double) decDeg;
    ms = ms*60.;
    decMin = (int) ms;
    decSec = (int) 60.*(ms - (double) decMin);
    sprintf(outputStr,"#:Sd +%02d:%02d:%02d#\n",decDeg,decMin,decSec);
  }
  else
  {
    dms = -newDec;
    dms = dms + 0.5/3600.;
    decDeg = (int) dms;
    ms = dms - (double) decDeg;
    ms = ms*60.;
    decMin = (int) ms;
    decSec = (int) 60.*(ms - (double) decMin);
    sprintf(outputStr,"#:Sd -%02d:%02d:%02d#\n",decDeg,decMin,decSec);
  }


  writen(TelPortFD,outputStr,16);
  if (readn(TelPortFD,inputStr,1,1) != 1 ) 
  {
    DisconnectTel();
    return -1;
  }

  if(inputStr[0] != '1') 
  {
    fprintf(stderr,"Error setting object Dec in SlewToCoords!\n");
    fprintf(stderr,"Tried to send command string %s\n",outputStr);
    fprintf(stderr,"Received string %s\n",inputStr);
  }

  writen(TelPortFD,"#:MS#",5);
  if (readn(TelPortFD,inputStr,1,5) != 1 ) 
  {
    DisconnectTel();
    return -1;
  }

  switch(inputStr[0]) 
  {
    case '0':
      return 0;
      break;
    case '1':
      returnVal = 1;
      fprintf(stderr,"Object is below horizon in SlewToCoords.\n");
      break;
    case '2':
      returnVal = 2;
      fprintf(stderr,"Declination is above upper limit in SlewToCoords.\n");
      break;
    default:
      fprintf(stderr,"Unexpected return from command #:MS#\n");
      return -1;
      break;
  }

  /* Look for '#' */
  
  for (;;) 
  {
    if ( readn(TelPortFD,inputStr,1,1) ) 
    {
      if (inputStr[0] == '#') break;
    }
    else 
    fprintf(stderr,"No acknowledgement from telescope after SlewToCoords.\n");
    return -1;
  }
  return returnVal;
}


int CheckCoords(double desRA, double desDec)
{
  double errorRA, errorDec, nowRA, nowDec;
        
  nowRA=GetRA();
  errorRA = nowRA - desRA;
  nowDec=GetDec();
  errorDec = nowDec - desDec;

  /* For 2 minute of arc precision */
  /*if( fabs(raError) > 1.0/(15.0*30.0) || fabs(decError) > 1.0/30.0)*/

  /* For 6 minute of arc precision */
  if( fabs(errorRA) > (0.1/15.) || fabs(errorDec) > 0.1)
    return 0;
  else
    return 1;
}


/* Set lower and upper limits to protect hardware */

void SetLimits(double limitLower, double limitHigher)
{
  int min, max;
  double dmin, dmax;
  char outputStr[32], inputStr[2048];

  min = (int) limitLower;
  max = (int) limitHigher;

  sprintf(outputStr,"#:So %02d#\n",min);
  writen(TelPortFD,outputStr,9);

  if (readn(TelPortFD,inputStr,1,1) != 1 ) 
  {
    fprintf(stderr,"Unable to set Dec lower limit\n");
  }  

  sprintf(outputStr,"#:Sh %02d#\n",max);
  writen(TelPortFD,outputStr,8);
  if (readn(TelPortFD,inputStr,1,1) != 1 ) 
  {
    fprintf(stderr,"Unable to set Dec upper limit\n");
  }  
}


static int writen(fd, ptr, nbytes)
int fd;
char *ptr;
int nbytes;
{
  int nleft, nwritten;
  nleft = nbytes;
  while (nleft > 0) 
  {
    nwritten = write (fd, ptr, nleft);
    if (nwritten <=0 ) break;
    nleft -= nwritten;
    ptr += nwritten;
  }
  return (nbytes - nleft);
}

static int readn(fd, ptr, nbytes, sec)
int fd;
char *ptr;
int nbytes;
int sec;
{
  int stat;
  int nleft, nread;
  nleft = nbytes;
  while (nleft > 0) 
  {
    stat = telstat(fd,sec,0);
    if (stat <=  0 ) break;
    nread  = read (fd, ptr, nleft);
    /* printf("readn: %d read\n", nread); */
    if (nread <= 0)  break;
    nleft -= nread;
    ptr += nread;
  }
  return (nbytes - nleft);
}

/*
 * Examines the read status of a file descriptor.
 * The timeout (sec, usec) specifies a maximum interval to
 * wait for data to be available in the descriptor.
 * To effect a poll, the timeout (sec, usec) should be 0.
 * Returns non-negative value on data available.
 * 0 indicates that the time limit referred by timeout expired.
 * On failure, it returns -1 and errno is set to indicate the
 * error.
 */
static int telstat(fd,sec,usec)
register int fd, sec, usec;
{
  int ret;
  int width;
  struct timeval timeout;
  telfds readfds;

  memset((char *)&readfds,0,sizeof(readfds));
  FD_SET(fd, &readfds);
  width = fd+1;
  timeout.tv_sec = sec;
  timeout.tv_usec = usec;
  ret = select(width,&readfds,NULL_PTR(telfds),NULL_PTR(telfds),&timeout);
  return(ret);
}


/* sleep scope, added ver.0.4.6 */
void sleepTel(void){
  int ret, flg = 1;
  char  returnStr[10];


  /* set target to az:0,alt:0 */
  writen(TelPortFD, "#:Sas00*00#", 11);
  writen(TelPortFD, "#:Sz000*00#", 11);

  /* point to az el */
  writen(TelPortFD, "#:MA#", 5);

  /* go to home position */
  //  #ifdef DEBUG
  //  printf("sending home command..\n");
  //  #endif
  //  writen(TelPortFD, "#:hF#", 5);
  sleep(15);

  //  while (flg) {
    //    #ifdef DEBUG
    //    printf("asking status..");
    //    #endif
    //    writen(TelPortFD, "#:h?#", 5);
    //    #ifdef DEBUG
    //    printf("done.\n");
    //    #endif
    //    ret = readn(TelPortFD, returnStr,1,1);
    //    #ifdef DEBUG
    //    //    printf("ret:%d %d\n", ret, returnStr[0]);
    //    #endif
    //    //    returnStr[ret]='\0';
    //    #ifdef DEBUG
    //    //    printf("returned %d: str:%s\n", returnStr);
    //    #endif
    //    if (returnStr[0] == '2') {
    //      /* in progress. wait 10sec. */
    //      sleep(10);
    //    } else {
    //      flg = 0;
    //    }
  //  }

  /* wait 5sec. */
  sleep(5);
  /* go to sleep */
  #ifdef DEBUG
  printf("sending sleep command..");
  #endif
  writen(TelPortFD, "#:hN#\n", 6);
  #ifdef DEBUG
  printf("done.\n");
  #endif

}

/* wake-up scope, added ver.0.4.6 */
void wakeTel(void) {
  writen(TelPortFD, "#:hW#", 5);
}

/* focus controls v0.5.0 20030911 */
void StartFocusPos(void) {
  /* set focus speed to slowest */
  writen(TelPortFD, "#:FS#", 5);

  writen(TelPortFD, "#:F+#", 5);
}

void StartFocusNeg(void) {
  /* set focus speed to slowest */
  writen(TelPortFD, "#:FS#", 5);

  writen(TelPortFD, "#:F-#", 5);
}

void StopFocus(void) {
  writen(TelPortFD, "#:FQ#", 5);
}

